DAT ProgramName         byte "Mfrc522Demo140623b", 0
CON
{{
  ****** Public Notes ******

  Code to demonstrate the object "Mfrc522_140321e".

  By: Duane Degn
  For: Dave Dorhout
  
  First attempt uploaded March 20, 2014.
  First working version uploaded March 21, 2014

  Version 140321e appears to read cards and tags
  correctly.

  The numbers listed in the "serialNumber" array
  down in the DAT section should be modified to
  match the numbers of your own cards and tags.

  The methods "PodBayDoorAction" and
  "SelfDestructAction" were added to the code
  as examples of using a code to start some
  sequence of code.
  
  For updated version of this code see the
  following thread in the Parallax Propeller
  Forum.

  http://forums.parallax.com/showthread.php/154851 

  See child object for additional credits.
  
}}
{
  ****** Private Notes ******

  Mf522Spi140319a Continue translating Arduino code
  to use sensor with Propeller.
  Change name to "Mfrc522Spi140319b".
  20c Still need a method to take the place of SPI.transfer.
  20c The code compiles. The program does not shift bits
  for SPI communication yet.
  20e SPI com needs to be msb first.
  20f Fix big CS problem with read.
  20h The local variable "backBits" wasn't
  being cleared before use in the
  "RequestFromMfrc522" method. Since only one byte of
  the variable was being written to, the other
  three bytes were interfering with the code.
  20h Appears to work correctly.
  20i Remove debug statements.
  20i Apparently I removed too much.
  20j Revert to h but use MainLoop from i.
  20j Appears to be working.
  21c Bug found in "Mfrc522ToCard" method. I needed to change
  a pair of "==" to "<>".
  21d Clean up code before posting to forum.
  21e Change name to "Mfrc522Demo140321e" in preparation
  to use this as a demo to the "Mfrc522_140321e" object.
  140321e Uploaded to forum.
  21f Use "Parallax Serial TerminalDat" to allow
  child object to send debug messages.
  140404a Change the structure of the program a bit.
  04a Use a top menu with several sub menus.
  04b Add password array to DAT section.
  140416a Doesn't read yet.
  140622a Added small demo to read from preset sectors.
  23b Both read and write works. Menu can only be used once
  between resets. I don't know why.
  23b There are two I/O pins used as triggers with a logic analyzer.
  Lots of clean up needed but it kind of works. 
  
}
CON

  _clkmode = xtal1 + pll16x 
  _xinfreq = 5_000_000

CON ' Pin Assignments

  SPI_CLK = 0
  SPI_MOSI = 4
  SPI_MISO = 2
  SPI_CS = 6
  RESET = 8
  INTERRUPT_PIN = 16
  DEBUG_PIN = 17
    
CON

  BAUD = 115_200
  QUOTE = 34
  END_OF_SERIAL_NUMBER_LIST = -1
  ATTEMPTED_READ_SIZE = 64
  MAX_READ_INDEX = ATTEMPTED_READ_SIZE - 1
  DUMPED_BUFFER_COLUMNS = 16
  TIMES_TO_CHECK_CARD_SIZE = 32
  BLOCK_SIZE = 16

  SECONDS_TILL_TIMEOUT = 5

  FIRST_READ_LOCATION = 16 '8
  LAST_READ_LOCATION = 23 '15
  ' cardIndex enumeration
  #0, POD_BAY_DOOR_ACTION, PASSWORD_ACTION_001
   
OBJ

  Pst : "Parallax Serial TerminalDat"
  Rfid : "Mfrc522_140623a"
  
VAR

  long cardNumber, cardIndex, maxIndex, activeCards
  long maxTextAddress, passwordIndex
  
  byte serNum[Rfid#MAX_LEN]
  byte tempBuffer[Rfid#MAX_LEN]   
  byte rxBuffer[ATTEMPTED_READ_SIZE + 2], readTrailerFlag
  
PUB Setup

  maxIndex := activeCards - 1
  maxTextAddress := @lastNumberText
  Pst.Start(BAUD)

  '' Comment out the five lines below to the have program immediately start.
  repeat
    Pst.Str(string(13, "Press any key to begin"))
    waitcnt(clkfreq / 4 + cnt)
    result := Pst.RxCount
  until result
  
  Pst.RxFlush

  Pst.Clear
  Pst.Home
  Pst.Str(string(13, "Program Name = ")) 
  Pst.Str(@ProgramName) 
  Pst.Str(string(13, "RFID Object Name = ")) 
  Pst.Str(Rfid.GetObjectName) 
  Pst.Str(string(13, "MFRC522 SPI Interface Demonstration Program", 13)) 
  Pst.Str(string(13, "Initializing RFID Reader", 13))

  DisplayTopMenu
 
  Rfid.Init(SPI_MOSI, SPI_MISO, SPI_CLK, SPI_CS, RESET)

  dira[DEBUG_PIN] := 1
  dira[Rfid#CHILD_LED] := 1 
  'DisplayReady
   
  TopLoop

PUB TopLoop 

  repeat
    result := Pst.RxCount
    if result
      result := Pst.CharIn
      case result
        "d", "D":
          \SimpleReadDemo(FIRST_READ_LOCATION, LAST_READ_LOCATION)
       { "a", "A":
          \ActionLoop
          '\CheckRfid
        "r", "R":
          \ReadLoop
        "w", "W":
          \WriteLoop    }
        "t", "T":
          \TestWrite(FIRST_READ_LOCATION, LAST_READ_LOCATION, @otherData)
          outa[DEBUG_PIN] := 0
        other:
          Pst.Str(string(13, "Not a vaild option."))
      DisplayTopMenu
      Pst.RxFlush
      
PUB ActionLoop

  DisplayActionMessage
  repeat
    result := Pst.RxCount
    if result
      result := Pst.CharIn
      case result
        '"w", "W":
          'WriteToCard()
          
        "x", "X":
          Pst.Str(string(13, "Exiting Action Menu"))
          Pst.Str(string(13, "Returning To Top Menu"))
          DisplayTopMenu
          abort
    CheckRfid

PUB SimpleReadDemo(firstReadLocation, lastReadLocation) | memoryIndex, startRead, endRead, localSector, localBlock, passwordBlock

  result := Rfid.RequestFromMfrc522(Rfid#PICC_REQIDL, @serNum)  
   
  if result == Rfid#MI_OK
    Pst.Str(string(13, "Card detected, type = $"))
    Pst.Hex(serNum[0], 2)
    Pst.Hex(serNum[1], 2)      
                                                     
    ' Anti-collision, return card serial number 4 bytes
    result := Rfid.Mfrc522Anticollision(@serNum)
   
    if result == Rfid#MI_OK
      result := Rfid.GetChecksum(@serNum)
      cardNumber := 0
      Rfid.SwapEndians(@cardNumber, @serNum)
     
      Pst.Str(string(13, "The card's number is  : $"))
      Pst.Hex(cardNumber, 8)
      Pst.Str(string(13, "Card's CRC value = $"))
      Pst.Hex(serNum[4], 2)
      Pst.Str(string(13, "Calculated checksum = $"))
      Pst.Hex(result, 2)
      if result & $FF == serNum[4]
        Pst.Str(string(13, "Card Data Successfully Read"))
      else
        Pst.Str(string(13, "Warning CRC Error!!!"))
        return
      Pst.Str(string(13, "Checking to see if card can be identified."))


      result := Rfid.Mfrc522SelectTag(@serNum)
      Pst.Str(string(13, "Card $")) 
      Pst.Hex(cardNumber, 8)   
      Pst.Str(string(" has a memory of "))
      Pst.Dec(result)
      Pst.Str(string(" units."))

      repeat memoryIndex from firstReadLocation to lastReadLocation
        Pst.Str(string(13, "memoryIndex = "))
        Pst.Dec(memoryIndex)
        Pst.Str(string(13, "Checking password"))
        passwordBlock := ComputePasswordBlock(memoryIndex)
        Pst.Str(string(13, "Password Block ")) 
        if memoryIndex == passwordBlock
          Pst.Str(string("Skipped (no data stored here)"))
          next
        else
          Pst.Str(string("= "))  
          Pst.Dec(passwordBlock) ' used to debug
        result := Rfid.MfrcAuthentication(Rfid#PICC_AUTHENT1A, passwordBlock, @keyA, @serNum)
         
        if result == Rfid#MI_OK
  
          result := Rfid.Mfrc522Read(memoryIndex, @rxBuffer) 
          if result == Rfid#MI_OK
            Pst.Str(string(13, "Successful Read"))  
            Pst.Str(string(13, "Data As Hex = "))  
            DumpBuffer(@rxBuffer, 16, 16)
            Pst.Str(string(13, "Data As Text = "))  
            SafeString(@rxBuffer, 16)
          else
            Pst.Str(string(13, "Mfrc522Read returned error = "))
            Pst.Dec(result)
        else
          Pst.Str(string(13, "MfrcAuthentication returned error = "))
          Pst.Dec(result)
        waitcnt(clkfreq / 10 + cnt)
         
    
      
      cardIndex := FindCardIndexNumber(cardNumber)
      if cardIndex == -1
        Pst.Str(string(13, "****** Unknown card ******"))
        Pst.Str(string(13, "You might want to write this number down in order"))
        Pst.Str(string(13, "to add it to the program's choices.", 13, "$"))
        Pst.Hex(cardNumber, 8)         
      else
        Pst.Str(string(13, "This is a known card."))
        Pst.Str(string(13, "The card's corresponding text is ", QUOTE))
        Pst.Str(FindStringOrderPtr(@correspondingText, cardIndex))
        Pst.Char(QUOTE)
        'Pst.Str(string(13, "Using the ", QUOTE, "case", QUOTE, " statement to branch"))
        'Pst.Str(string(13, "the program based on which card was read."))
                 
        case cardIndex
          POD_BAY_DOOR_ACTION:
            PasswordAction(@serNum)  
            'PodBayDoorAction
          PASSWORD_ACTION_001:
            PasswordAction(@serNum)
          other:                                
            Pst.Str(string(13, "This card was recognized but no action has been "))
            Pst.Str(string(" associated with it."))
     
    else
      Pst.Str(string(13, "There was a problem reading from the card."))
      Pst.Str(string(13, "The returned error code = "))
      Pst.Dec(result)
      Pst.Str(string(13, "The data read = $"))
      repeat result from 0 to 4
        Pst.Hex(serNum[result], 2)
        if result <> 4
          Pst.Str(string(", $"))
          
    waitcnt(clkfreq / 2 + cnt)
    
    DisplayReady
   
  else
    Pst.Str(string(13, "SimpleReadDemo fail result = "))
    Pst.Dec(result)
    result := 0
         
PUB ComputePasswordBlock(memoryIndex)

  result := memoryIndex // 4
  case result
    3:
      result := memoryIndex 
    other:
      result := memoryIndex + 3 - result

PUB SimpleDemoOld | memoryIndex, startRead, endRead, localSector, localBlock

  startRead := 8
  endRead := 15
  repeat memoryIndex from startRead to endRead
    ReadAttempt(memoryIndex, @keyA, @rxBuffer)
    SafeString(@rxBuffer, 16)
    {localSector := ComputeSector(memoryIndex)
    localBlock := ComputeBlock(memoryIndex)
    Pst.Str(string(13, "Attempting to read from location #"))
    Pst.Dec(memoryIndex)
    Pst.Str(string(", sector #"))
    Pst.Dec(localSector)
    Pst.Str(string(", block #"))
    Pst.Dec(localBlock)       }
    
         
PUB ComputeSector(memoryIndex)

  result := memoryIndex / 4

PUB ComputeBlock(memoryIndex)

  result := memoryIndex // 4

PUB ReadLoop | sector, block, size

  longfill(@sector, -1, 2)
  size := 1
  DisplayReadMenu(sector, block, size) 
  repeat
    result := Pst.RxCount
    if result
      result := Pst.CharIn
      case result
        "r", "R":
          ReadFromCard(sector, block, size)
        "s", "S": 'sector
          sector := GetSector
          DisplayReadMenu(sector, block, size) 
        "b", "B": 'block
          block := GetBlock
          DisplayReadMenu(sector, block, size) 
        '"p", "P": 'block
        '  passwordIndex := GetPasswordIndex
        '  DisplayReadMenu(sector, block, size) 
        "t", "T": 'toggle
          if readTrailerFlag
            readTrailerFlag := 0
          else
            readTrailerFlag := 1  
          DisplayReadMenu(sector, block, size) 
        "n", "N": 'size
          size := GetSize
          DisplayReadMenu(sector, block, size) 
        "x", "X":
          Pst.Str(string(13, "Exiting Read Menu"))
          Pst.Str(string(13, "Returning To Top Menu"))
          DisplayTopMenu
          abort

PUB WriteLoop | sector, block, size

  longfill(@sector, -1, 2)
  size := 1
  DisplayWriteMenu(sector, block, size) 
  repeat
    result := Pst.RxCount
    if result
      result := Pst.CharIn
      case result
        "w", "W":
          WriteToCard(sector, block, size)
        "s", "S": 'sector
          sector := GetSector
          DisplayWriteMenu(sector, block, size)
        "b", "B": 'block
          block := GetBlock
          DisplayWriteMenu(sector, block, size)
        "t", "T": 'toggle
          if readTrailerFlag
            readTrailerFlag := 0
          else
            readTrailerFlag := 1  
          DisplayWriteMenu(sector, block, size) 
        "n", "N": 'size
          size := GetSize
          DisplayReadMenu(sector, block, size) 
        "x", "X":
          Pst.Str(string(13, "Exiting Write Menu"))
          Pst.Str(string(13, "Returning To Top Menu"))
          DisplayTopMenu
          abort
         
PUB TestWrite(firstWriteLocation, lastWriteLocation, writeBuffer) | memoryIndex, startRead, endRead, localSector, localBlock, passwordBlock

  Pst.Str(string(13, "TestWrite"))
  outa[DEBUG_PIN] := 1
  result := Rfid.RequestFromMfrc522(Rfid#PICC_REQIDL, @serNum)  
   
  if result == Rfid#MI_OK
    Pst.Str(string(13, "Card detected, type = $"))
    Pst.Hex(serNum[0], 2)
    Pst.Hex(serNum[1], 2)      
                                                     
    ' Anti-collision, return card serial number 4 bytes
    result := Rfid.Mfrc522Anticollision(@serNum)
   
    if result == Rfid#MI_OK
      result := Rfid.GetChecksum(@serNum)
      cardNumber := 0
      Rfid.SwapEndians(@cardNumber, @serNum)
     
      Pst.Str(string(13, "The card's number is  : $"))
      Pst.Hex(cardNumber, 8)
      Pst.Str(string(13, "Card's CRC value = $"))
      Pst.Hex(serNum[4], 2)
      Pst.Str(string(13, "Calculated checksum = $"))
      Pst.Hex(result, 2)
      if result & $FF == serNum[4]
        Pst.Str(string(13, "Card Data Successfully Read"))
      else
        Pst.Str(string(13, "Warning CRC Error!!!"))
        return
      Pst.Str(string(13, "Checking to see if card can be identified."))


      result := Rfid.Mfrc522SelectTag(@serNum)
      Pst.Str(string(13, "Card $")) 
      Pst.Hex(cardNumber, 8)   
      Pst.Str(string(" has a memory of "))
      Pst.Dec(result)
      Pst.Str(string(" units."))

      repeat memoryIndex from firstWriteLocation to lastWriteLocation
        Pst.Str(string(13, "memoryIndex = "))
        Pst.Dec(memoryIndex)
        Pst.Str(string(13, "Checking password"))
        passwordBlock := ComputePasswordBlock(memoryIndex)
        Pst.Str(string(13, "Password Block ")) 
        if memoryIndex == passwordBlock
          Pst.Str(string("Skipped (no data to be stored here)"))
          next
        else
          Pst.Str(string("= "))  
          Pst.Dec(passwordBlock) ' used to debug
        result := Rfid.MfrcAuthentication(Rfid#PICC_AUTHENT1A, passwordBlock, @keyA, @serNum)
         
        if result == Rfid#MI_OK
   
          result := Rfid.Mfrc522Write(memoryIndex, writeBuffer)
          
          if result == Rfid#MI_OK
            Pst.Str(string(13, "Successful Write"))
            Pst.Str(string(13, "Written to card = "))  
            SafeString(writeBuffer, 16)
          else
            Pst.Str(string(13, "Mfrc522Write returned error = "))
            Pst.Dec(result)
          writeBuffer += 16   
        else
          Pst.Str(string(13, "MfrcAuthentication returned error = "))
          Pst.Dec(result)
        waitcnt(clkfreq / 10 + cnt)
     
    else
      Pst.Str(string(13, "There was a problem reading from the card."))
      Pst.Str(string(13, "The returned error code = "))
      Pst.Dec(result)
      Pst.Str(string(13, "The data read = $"))
      repeat result from 0 to 4
        Pst.Hex(serNum[result], 2)
        if result <> 4
          Pst.Str(string(", $"))
          
    waitcnt(clkfreq / 2 + cnt)
    
    DisplayReady
   
  else
    Pst.Str(string(13, "TestWrite fail result = "))
    Pst.Dec(result)
    result := 0
         
PUB TestLoopOld

  DisplayTestMenu 
  repeat
    result := Pst.RxCount
    if result
      result := Pst.CharIn
      case result
        "i":
          MonitorInterruptRegister
          Pst.RxFlush
          return
        "1":
          Rfid.WriteMfrc522(Rfid#CommIrqReg + 1, %1001_0000)  ' Turn on IRQ pin?
        "v":
          ReadChipVersion
        "s":
          FindSizeOfActiveCard(@serNum)
        
        "x", "X":
          Pst.Str(string(13, "Exiting Test Menu"))
          Pst.Str(string(13, "Returning To Top Menu"))
          DisplayTopMenu
          abort
         
PUB FindSizeOfActiveCard(serialNumberPtr)
  
  if byte[serialNumberPtr] or byte[serialNumberPtr][1] or byte[serialNumberPtr][2] or {
    } byte[serialNumberPtr][3] ' if serial number isn't zero
    result := Rfid.Mfrc522SelectTag(serialNumberPtr)
    Pst.Str(string(13, "Card $")) 
    Pst.Hex(cardNumber, 8)   
    Pst.Str(string(" has a memory of "))
    Pst.Dec(result)
    Pst.Str(string(" units."))
  else
    Pst.Str(string(13, "No active card.")) 
    Pst.Str(string(13, "Please hold card next to reader and try again.")) 
    result := -1
    
PUB ReadFromCard(sector, block, size) | timer, timeoutInterval

  timeoutInterval := clkfreq * SECONDS_TILL_TIMEOUT
  
  block += sector * 4
  Pst.Str(string(13, "Preparing to read from card."))
  Pst.Str(string(13, "Blocks to be write = "))  
  Pst.Dec(size) 
  Pst.Str(string(13, "readTrailerFlag = "))
  Pst.Dec(readTrailerFlag)
  
  Pst.Str(string(13, "Place card near reader."))
  
  timer := cnt
  repeat
    result := CheckForCard
  while result == 0 and cnt - timer < timeoutInterval

  if result == 0
    Pst.Str(string(13, "Read from card timeout."))
    return 0
  result := 0  
  repeat size
    if block // 4 == 3 and readTrailerFlag == 0
      block++ 
    Pst.Str(string(13, "Sector = "))
    Pst.Dec(block / 4)    
    Pst.Str(string(13, "Block = "))
    Pst.Dec(block // 4) 
    result += ReadAttempt(block++, @keyA, @rxBuffer)
    Pst.Str(string(13, "data from card = "))
    SafeString(@rxBuffer, 16)
  Pst.Str(string(13, "Successful Reads = "))
  Pst.Dec(result)

PUB ReadAttempt(blockAddress, keyPtr, rxPtr) | passwordBlock 

  passwordBlock := blockAddress - (blockAddress // 4) + 3

  Pst.Str(string(13, "Key A = "))
  HexDebug(keyPtr, 6) 
  Pst.Str(string(13, "Sector = "))
  Pst.Dec(blockAddress / 4)    
  Pst.Str(string(13, "Block = "))
  Pst.Dec(blockAddress // 4)
  
  'status = MFRC522_Auth(PICC_AUTHENT1A, passwordBlock, key, serNum);      //Authenticate
  result := Rfid.MfrcAuthentication(Rfid#PICC_AUTHENT1A, passwordBlock, keyPtr, @serNum) 
  if result == Rfid#MI_OK

    'status = MFRC522_Read(blockAddr, str);
    result := Rfid.Mfrc522Read(blockAddress, rxPtr)
    if result == Rfid#MI_OK
      Pst.Str(string(13, "Read from the card, the data is : "))
      SafeString(rxPtr, 16)
     
    else
      Pst.Str(string(13, "Mfrc522Read error # "))
      Pst.Dec(result)
  else
    Pst.Str(string(13, "MfrcAuthentication error # "))
    Pst.Dec(result)

PUB WriteToCard(sector, block, size)  | timer, timeoutInterval, dataIndex

  timeoutInterval := clkfreq * SECONDS_TILL_TIMEOUT

  Pst.Str(string(13, "Preparing to write to card."))
  Pst.Str(string(13, "Sector = "))
  Pst.Dec(sector)    
  Pst.Str(string(13, "Block = "))
  Pst.Dec(block)
  Pst.Str(string(13, "Blocks to be write = "))  
  Pst.Dec(size) 
  
  if sector < 0 or sector > 15
    Pst.Str(string(13, "Not a legitimate sector. End write."))
    return 

  Pst.Str(string(13, "Place card near reader."))

  timer := cnt
  repeat
    result := CheckForCard
  while result == 0 and cnt - timer < timeoutInterval
  
  dataIndex := 0
  block += sector * 4
  repeat size
    if block == 0
      Pst.Str(string(13, "Block 0 is read only."))
      block++
    if block // 4 == 3
      block++
    if dataIndex == 9
      dataIndex := 0
    else
      dataIndex++  
    'Write(i, writeData[dataIndex], sectorKeyA[0], sectorNewKeyA[1]);  
    result += WriteAttempt(block++, @otherData + (dataIndex * 16), @keyA, @newKey)

  Pst.Str(string(13, "Successful Writes = "))
  Pst.Dec(result)

           
PUB WriteAttempt(blockAddress, dataPtr, keyPtr, newKeyPtr) | successCount, keyAddress, passwordBlock

  passwordBlock := blockAddress - (blockAddress // 4) + 3

  Pst.Str(string(13, "Key A = "))
  HexDebug(keyPtr, 6)
  Pst.Str(string(13, "New Key = "))
  HexDebug(newKeyPtr, 16)
  Pst.Str(string(13, "Data to write = "))
  SafeString(dataPtr, 16)
  Pst.Str(string(13, "Sector = "))
  Pst.Dec(blockAddress / 4)    
  Pst.Str(string(13, "Block = "))
  Pst.Dec(blockAddress // 4)
  'status = MFRC522_Auth(PICC_AUTHENT1A, passwordBlock, key, serNum);  
  result := Rfid.MfrcAuthentication(Rfid#PICC_AUTHENT1A, passwordBlock, keyPtr, @serNum) 
  if result == Rfid#MI_OK
    'status = MFRC522_Write(passwordBlock, newkey);
    result := Rfid.Mfrc522Write(passwordBlock, newKeyPtr)
    'status = MFRC522_Write(blockAddr, data);    
    result := Rfid.Mfrc522Write(blockAddress, dataPtr) 
    if result == Rfid#MI_OK
      Pst.Str(string(13, "Successful Write"))
      result := 1
    else
      Pst.Str(string(13, "Mfrc522Write returned error = "))
      Pst.Dec(result)
      result := 0
  else
    Pst.Str(string(13, "MfrcAuthentication returned error = "))
    Pst.Dec(result)
    result := 0 
  'waitcnt(clkfreq / 10 + cnt)
      
PUB MonitorInterruptRegister | tempValue

  repeat
    repeat result from Rfid#CommIrqReg to Rfid#CommIrqReg + 2
      tempValue := Rfid.ReadMfrc522(result)
      Pst.Str(string(13, "CommIrqReg["))
      Pst.Dec(result - Rfid#CommIrqReg)
      Pst.Str(string("] = $"))
      Pst.Hex(tempValue, 8)
      Pst.Str(string(" or "))
      ReadableBin(tempValue, 8)
      Pst.Str(string(13, "Interrupt Pin = "))
      Pst.Dec(ina[INTERRUPT_PIN])
    result := Pst.RxCount
  until result
  
PUB ReadChipVersion

  result := Rfid.ReadMfrc522(Rfid#VersionReg)
  Pst.Str(string(13, "Returned value from chip version register = $"))
  Pst.Hex(result, 2)
  case result
    $91:
      Pst.Str(string(13, "The value $91 indicates this chip uses software version # 1.")) 
    $92:
      Pst.Str(string(13, "The value $92 indicates this chip uses software version # 2."))   
    other:
      Pst.Str(string(13, "This is strange. The expected return values were $91 or $92."))
      Pst.Str(string(13, 13, "****** Unknown software version. ******", 13))
  Pst.Char(13)
  Pst.Char(13)
  
PUB CheckRfid | blockAddress
  
  result := Rfid.RequestFromMfrc522(Rfid#PICC_REQIDL, @serNum)  
   
  if result == Rfid#MI_OK
    Pst.Str(string(13, "Card detected, type = $"))
    Pst.Hex(serNum[0], 2)
    Pst.Hex(serNum[1], 2)      
                                                     
    ' Anti-collision, return card serial number 4 bytes
    result := Rfid.Mfrc522Anticollision(@serNum)
   
    if result == Rfid#MI_OK
      result := Rfid.GetChecksum(@serNum)
      cardNumber := 0
      Rfid.SwapEndians(@cardNumber, @serNum)
     
      Pst.Str(string(13, "The card's number is  : $"))
      Pst.Hex(cardNumber, 8)
      Pst.Str(string(13, "Card's CRC value = $"))
      Pst.Hex(serNum[4], 2)
      Pst.Str(string(13, "Calculated checksum = $"))
      Pst.Hex(result, 2)
      if result & $FF == serNum[4]
        Pst.Str(string(13, "Card Data Successfully Read"))
      else
        Pst.Str(string(13, "Warning CRC Error!!!"))
        return
      Pst.Str(string(13, "Checking to see if card can be identified."))


      result := Rfid.Mfrc522SelectTag(@serNum)
      Pst.Str(string(13, "Card $")) 
      Pst.Hex(cardNumber, 8)   
      Pst.Str(string(" has a memory of "))
      Pst.Dec(result)
      Pst.Str(string(" units."))
      blockAddress := 7
      'result := Rfid.MfrcAuthentication(Rfid#PICC_AUTHENT1A, blockAddress, @sectorKeyA + (blockAddress / 4), @serNum)
      Pst.Str(string(13, "Maybe checking password???"))  
      result := Rfid.MfrcAuthentication(Rfid#PICC_AUTHENT1A, blockAddress, @sectorNewKeyA + (blockAddress / 4), @serNum)
       
      if result == Rfid#MI_OK
        blockAddress -= 3
        result := Rfid.Mfrc522Read(blockAddress, @rxBuffer) 
        if result == Rfid#MI_OK
          Pst.Str(string(13, "Successful Read"))  
          Pst.Str(string(13, "Data As Hex = "))  
          DumpBuffer(@rxBuffer, 16, 16)
          Pst.Str(string(13, "Data As Text = "))  
          SafeString(@rxBuffer, 16)
        else
          Pst.Str(string(13, "Mfrc522Read returned error = "))
          Pst.Dec(result)
      else
        Pst.Str(string(13, "MfrcAuthentication returned error = "))
        Pst.Dec(result)
      waitcnt(clkfreq / 10 + cnt)
  
    
      
      cardIndex := FindCardIndexNumber(cardNumber)
      if cardIndex == -1
        Pst.Str(string(13, "****** Unknown card ******"))
        Pst.Str(string(13, "You might want to write this number down in order"))
        Pst.Str(string(13, "to add it to the program's choices.", 13, "$"))
        Pst.Hex(cardNumber, 8)         
      else
        Pst.Str(string(13, "This is a known card."))
        Pst.Str(string(13, "The card's corresponding text is ", QUOTE))
        Pst.Str(FindStringOrderPtr(@correspondingText, cardIndex))
        Pst.Char(QUOTE)
        'Pst.Str(string(13, "Using the ", QUOTE, "case", QUOTE, " statement to branch"))
        'Pst.Str(string(13, "the program based on which card was read."))
                 
        case cardIndex
          POD_BAY_DOOR_ACTION:
            PasswordAction(@serNum)  
            'PodBayDoorAction
          PASSWORD_ACTION_001:
            PasswordAction(@serNum)
          other:                                
            Pst.Str(string(13, "This card was recognized but no action has been "))
            Pst.Str(string(" associated with it."))
     
    else
      Pst.Str(string(13, "There was a problem reading from the card."))
      Pst.Str(string(13, "The returned error code = "))
      Pst.Dec(result)
      Pst.Str(string(13, "The data read = $"))
      repeat result from 0 to 4
        Pst.Hex(serNum[result], 2)
        if result <> 4
          Pst.Str(string(", $"))
          
    waitcnt(clkfreq / 2 + cnt)
    
    DisplayReady
   
  else
    result := 0
        
PUB CheckForCard | blockAddress
  
  result := Rfid.RequestFromMfrc522(Rfid#PICC_REQIDL, @serNum)  
   
  if result == Rfid#MI_OK
    Pst.Str(string(13, "Card detected, type = $"))
    Pst.Hex(serNum[0], 2)
    Pst.Hex(serNum[1], 2)      
                                                     
    ' Anti-collision, return card serial number 4 bytes
    result := Rfid.Mfrc522Anticollision(@serNum)
   
    if result == Rfid#MI_OK
      result := Rfid.GetChecksum(@serNum)
      cardNumber := 0
      Rfid.SwapEndians(@cardNumber, @serNum)
     
      Pst.Str(string(13, "The card's number is  : $"))
      Pst.Hex(cardNumber, 8)
      Pst.Str(string(13, "Card's CRC value = $"))
      Pst.Hex(serNum[4], 2)
      Pst.Str(string(13, "Calculated checksum = $"))
      Pst.Hex(result, 2)
      if result & $FF == serNum[4]
        Pst.Str(string(13, "Card Data Successfully Read"))
      else
        Pst.Str(string(13, "Warning CRC Error!!!"))
        return 0
      Pst.Str(string(13, "Checking to see if card can be identified."))
      
      result := Rfid.Mfrc522SelectTag(@serNum)
      Pst.Str(string(13, "Card $")) 
      Pst.Hex(cardNumber, 8)   
      Pst.Str(string(" has a memory of "))
      Pst.Dec(result)
      Pst.Str(string(" units."))
     
    else
      Pst.Str(string(13, "There was a problem reading from the card."))
      Pst.Str(string(13, "The returned error code = "))
      Pst.Dec(result)
      Pst.Str(string(13, "The data read = $"))
      repeat result from 0 to 4
        Pst.Hex(serNum[result], 2)
        if result <> 4
          Pst.Str(string(", $"))
      result := 0
  else
    result := 0
        
PUB GetSector

  Pst.Str(string(13, "Please enter sector."))
  result := Pst.DecIn
  
PUB GetBlock

  Pst.Str(string(13, "Please enter block."))
  result := Pst.DecIn
  
PUB GetSize

  Pst.Str(string(13, "Please enter number of blocks to read or write."))
  result := Pst.DecIn


PUB GetPasswordIndex

  Pst.Str(string(13, "Please enter password index."))
  result := Pst.DecIn
  
PUB DisplayReady

  Pst.Str(string(13, 13, 13, "Ready to read card."))    

PUB PodBayDoorAction

  Pst.Str(string(13, "Open pod bay doors method."))
  'Pst.Char(7)
  waitcnt(clkfreq / 2 + cnt)
  'Pst.Char(7)
  waitcnt(clkfreq / 2 + cnt)
  'Pst.Char(7)
  waitcnt(clkfreq / 2 + cnt)  
  Pst.Str(string(13, "Sorry Dave I can't do that."))
  
PUB PasswordAction(serialNumberPtr) | blockAddress
'' Guessing how this is done based on Arduino code.  

  Pst.Str(string(13, "Checking data size of card."))
  repeat TIMES_TO_CHECK_CARD_SIZE
    result := Rfid.Mfrc522SelectTag(serialNumberPtr)
    if result
      quit
  Pst.Str(string(13, "Card $")) 
  Pst.Hex(cardNumber, 8)        ' This assumes the number at "serialNumberPtr" is the same as "cardNumber".
                                ' Not a clean way to do this.
  Pst.Str(string(" has a memory of "))
  Pst.Dec(result)
  Pst.Str(string(" units."))
  'ifnot result
    'Pst.Str(string("This action requires a card with memory. Please use with different card."))
    'return
  'Pst.Str(string(13, "Reading from to block address $07 since that's what the example did."))
  blockAddress := 7
  result := Rfid.MfrcAuthentication(Rfid#PICC_AUTHENT1A, blockAddress, @sectorKeyA + (blockAddress / 4), serialNumberPtr)
  Pst.Str(string(13, "Maybe checking password???"))  
  'result := Rfid.MfrcAuthentication(Rfid#PICC_AUTHENT1A, blockAddress, @sectorNewKeyA + (blockAddress / 4), serialNumberPtr)

  if result == Rfid#MI_OK
    blockAddress -= 3
    result := Rfid.Mfrc522Read(blockAddress, @rxBuffer) 
    if result == Rfid#MI_OK
      Pst.Str(string(13, "Successful Read"))  
      Pst.Str(string(13, "Data As Hex = "))  
      DumpBuffer(@rxBuffer, 16, 16)
      Pst.Str(string(13, "Data As Text = "))  
      SafeString(@rxBuffer, 16)
    else
      Pst.Str(string(13, "Mfrc522Read returned error = "))
      Pst.Dec(result)
  else
    Pst.Str(string(13, "MfrcAuthentication returned error = "))
    Pst.Dec(result)
  waitcnt(clkfreq / 10 + cnt)
   
PUB ReadableBin(localValue, localSize)

  localSize += 3
  localSize -= localSize // 4
  Pst.char("%")
  'result := localValue
  localValue <-= 32 - localSize
  repeat while localSize
    localValue <-= 4
    Pst.bin(localValue, 4)
    localSize -= 4
    if localSize
      Pst.char("_")

PUB FindStringOrderPtr(firstStr, stringIndex)     
'' Finds start address of one string in a list
'' of string. "firstStr" is the address of 
'' string #0 in the list. "stringIndex"
'' indicates which of the strings in the list
'' the method is to find.

  result := firstStr
  
  repeat while stringIndex    
    repeat while byte[result++]  
    stringIndex--


PUB FindCardIndexNumber(serialNumberToMatch)     
'' Finds the index number of card's serial number.
'' Used to find information associated with the
'' card.

  result := 0

  repeat while serialNumber[result] <> serialNumberToMatch and {
    } serialNumber[result] <> END_OF_SERIAL_NUMBER_LIST
    result++
  if serialNumber[result] == END_OF_SERIAL_NUMBER_LIST
    result := -1

PUB FindCardIndexAndText(serialNumberToMatch)
'' Displays text from DAT section corresponding to the
'' appropriate card.
'' Returns the card's index number (an easier to manage number
'' than the large serial numbers).
'' This method is not presently used.

  result := FindCardIndexNumber(serialNumberToMatch)
  if result == -1 
    Pst.Str(@notFoundText)
  else
    Pst.Str(FindStringOrderPtr(@correspondingText, result))
       
PUB DumpBuffer(localPtr, localSize, localColumns) 

  localSize += localPtr
  
  repeat  
    Pst.Char(13)     
    repeat localColumns
      Pst.Char("$")
      Pst.Hex(byte[localPtr++], 2)
      Pst.Str(string(",  "))
      if localPtr > localSize
        return
          
PUB SafeString(localPtr, localSize)

  repeat localSize
    case byte[localPtr]
      32..127:
        Pst.Char(byte[localPtr])
      other:
        Pst.Char("<")
        Pst.Char("$")
        Pst.Hex(byte[localPtr], 2)
        Pst.Char(">")
    localPtr++
    
PUB HexDebug(localPtr, localSize)

  repeat localSize
    Pst.Char("$")
    Pst.Hex(byte[localPtr], 2)
    Pst.Char(" ")
    localPtr++
    
PUB DisplayTopMenu

  Pst.Str(string(13, "     Top Menu ", 13))       
  Pst.Str(string(13, "Please Select Which Submenu to Use"))
  Pst.Str(string(13, "Use ", QUOTE, "X", QUOTE, " to return to this top menu.")) 

 { Pst.Str(string(13, "A) take Action based on which card is found"))   
  Pst.Str(string(13, "R) Read from card"))   
  Pst.Str(string(13, "W) Write to card"))  } 
  Pst.Str(string(13, "T) Test new features"))   
  Pst.Str(string(13, "D) simple Demo (programmed test stuff)"))   
 
PUB DisplayActionMessage

  Pst.Str(string(13, "     Action Mode", 13))       
  Pst.Str(string(13, "Please Select Read Action"))
    
  Pst.Str(string(13, "X) eXit to top menu")) 

PUB DisplayReadMenu(sector, block, size)

  Pst.Str(string(13, "     Read Menu ", 13))

  if sector == -1 and block == -1      
    Pst.Str(string(13, "The sector and block numbers need to be entered prior to reading from card."))
  elseif sector == -1
    Pst.Str(string(13, "Block to be read = "))
    Pst.Dec(block)    
    Pst.Str(string(13, "The sector numbers needs to be entered prior to reading from card."))
  elseif block == -1      
    Pst.Str(string(13, "Sector to be written = "))
    Pst.Dec(sector)    
    Pst.Str(string(13, "The block number needs to be entered prior to reading from card."))
  else
    Pst.Str(string(13, "Sector = "))
    Pst.Dec(sector)    
    Pst.Str(string(13, "Block = "))
    Pst.Dec(block)   
    Pst.Str(string(13, "The sector and block numbers may be changed prior to reading from card."))
    Pst.Str(string(13, "To read from the card with the present sector and block values press ", QUOTE, "R", QUOTE, "."))
  'Pst.Str(string(13, "passwordIndex = "))
  'Pst.Dec(passwordIndex)
  Pst.Str(string(13, "active password = "))
  DisplayHexBytes(@keyA + (BYTES_PER_KEY * passwordIndex), BYTES_PER_KEY)
 
  Pst.Str(string(13, "Blocks to be read = "))  
  Pst.Dec(size) 
  Pst.Str(string(13, "readTrailerFlag = "))
  Pst.Dec(readTrailerFlag) 
  Pst.Str(string(13, "Please Select Read Action"))
  if sector <> -1 and block <> -1
    'Pst.Str(string(13, "R) Read from card using present sector, block and password settings"))   
    Pst.Str(string(13, "R) Read from card using present sector and block settings"))   
   
  Pst.Str(string(13, "S) enter Sector to be read"))   
  Pst.Str(string(13, "B) enter Block to be read")) 
  'Pst.Str(string(13, "P) enter Password index (password not currently used)")) 
  Pst.Str(string(13, "T) toggle read Trailer flag")) 
  Pst.Str(string(13, "N) enter Number of blocks to read")) 
  Pst.Str(string(13, "X) eXit to top menu")) 

PUB DisplayHexBytes(localPtr, bytesToDisplay)

  bytesToDisplay-- 
  repeat result from 0 to bytesToDisplay
    Pst.Char("$")
    Pst.Hex(byte[localPtr][result], 2)
    if result <> bytesToDisplay
      Pst.Char(",")
      Pst.Char(" ")
    
PUB DisplayWriteMenu(sector, block, size)

  Pst.Str(string(13, "     Write Menu ", 13))
  if sector == -1 and block == -1      
    Pst.Str(string(13, "The sector and block numbers need to be entered prior to writing to card."))
  elseif sector == -1
    Pst.Str(string(13, "Block to be written = "))
    Pst.Dec(block)    
    Pst.Str(string(13, "The sector numbers needs to be entered prior to writing to card."))
  elseif block == -1      
    Pst.Str(string(13, "Sector to be written = "))
    Pst.Dec(sector)    
    Pst.Str(string(13, "The block number needs to be entered prior to writing to card."))
  else
    Pst.Str(string(13, "Sector = "))
    Pst.Dec(sector)    
    Pst.Str(string(13, "Block = "))
    Pst.Dec(block)   
    Pst.Str(string(13, "The sector and block numbers may be changed prior to writing to card."))
    Pst.Str(string(13, "The write to the card with the present sector and block values press ", QUOTE, "W", QUOTE, "."))
  Pst.Str(string(13, "Blocks to be read = "))  
  Pst.Dec(size) 
  Pst.Str(string(13, "Please Select Write Action"))
  if sector <> -1 and block <> -1
    Pst.Str(string(13, "W) Write to card with present sector and block settings"))   
  Pst.Str(string(13, "S) enter Sector to write"))   
  Pst.Str(string(13, "B) enter block to write")) 
  Pst.Str(string(13, "N) enter Number of blocks to write")) 
  Pst.Str(string(13, "X) eXit to top menu")) 

PUB DisplayTestMenu

  Pst.Str(string(13, "     Test Menu ", 13))       
  Pst.Str(string(13, "Please Select Test Action"))
  Pst.Str(string(13, "S) Serial number"))   
  'Pst.Str(string(13, "P) enter Password for block (password not currently used)"))   
  Pst.Str(string(13, "T) Test new features"))   
  Pst.Str(string(13, "X) eXit to top menu")) 
 
PRI BuildAccessBit(destPtr, accessNibble1, accessNibble2, accessNibble3)


  byte[destPtr++] := (!accessNibble2 << 4) |  !accessNibble1
  byte[destPtr++] := (accessNibble1 << 4) |  !accessNibble3
  byte[destPtr++] := (accessNibble3 << 4) |  accessNibble2
  
CON

  MAX_POSSIBLE_CARDS = 16
  NUMBER_OF_HARD_CODED_CARDS = 2
  AVERAGE_TEXT_SIZE = 16
  CARDS_IN_USE = 2
  NUMBER_OF_SECTORS = 16
  MAX_SECTOR_INDEX = NUMBER_OF_SECTORS - 1
  NUMBER_OF_BLOCKS_PER_SECTOR = 4
  MAX_BLOCK_INDEX = NUMBER_OF_BLOCKS_PER_SECTOR - 1
  MAX_USEABLE_BLOCK_INDEX = MAX_BLOCK_INDEX - 1
  BYTES_PER_KEY = 6
 { READ_A_B_WRITE_B = %100
  RW_ACCESS_BITS_C123_BIT_0 = %100
  RW_ACCESS_BITS_C123_BIT_1 = %100
  RW_ACCESS_BITS_C123_BIT_2 = %100
  RW_ACCESS_BITS_C123_BIT_3 = %011
  RO_ACCESS_BITS_C123_BIT_0 = %010
  RO_ACCESS_BITS_C123_BIT_1 = %010
  RO_ACCESS_BITS_C123_BIT_2 = %010
  RO_ACCESS_BITS_C123_BIT_3 = %110  }
  RW_ACCESS_BITS_C1 = %0111 '1110
  RW_ACCESS_BITS_C2 = %1000 '0001
  RW_ACCESS_BITS_C3 = %1000 '0001
  
  RO_ACCESS_BITS_C1 = %1000 '0001
  RO_ACCESS_BITS_C2 = %1111
  RO_ACCESS_BITS_C3 = %0000
                              '!C2 !C1
  RW_SECTOR_TRAILER_BYTE_6 = %0111_1000 
                              ' C1 !C3
  RW_SECTOR_TRAILER_BYTE_7 = %0111_0111 
                              ' C3  C2
  RW_SECTOR_TRAILER_BYTE_8 = %1000_1000
  
                              '!C2 !C1
  RO_SECTOR_TRAILER_BYTE_6 = %0000_0111 
                              ' C1 !C3
  RO_SECTOR_TRAILER_BYTE_7 = %1000_1111 
                              ' C3  C2
  RO_SECTOR_TRAILER_BYTE_8 = %0000_1111 


  {RW_NFC_BITS_C123_BIT_0 = %000
  RW_NFC_BITS_C123_BIT_1 = %000
  RW_NFC_BITS_C123_BIT_2 = %000
  RW_NFC_BITS_C123_BIT_3 = %011
  
  RO_NFC_BITS_C123_BIT_0 = %010
  RO_NFC_BITS_C123_BIT_1 = %010
  RO_NFC_BITS_C123_BIT_2 = %010
  RO_NFC_BITS_C123_BIT_3 = %110   }

  RW_NFC_BITS_C1 = %0000
  RW_NFC_BITS_C2 = %1000 
  RW_NFC_BITS_C3 = %1000 
  
  RO_NFC_BITS_C1 = %1000
  RO_NFC_BITS_C2 = %1111
  RO_NFC_BITS_C3 = %0000   
                           '!C2 !C1
  RW_NFC_TRAILER_BYTE_6 = %0111_1111 
                           ' C1 !C3
  RW_NFC_TRAILER_BYTE_7 = %0000_0111 
                           ' C3  C2
  RW_NFC_TRAILER_BYTE_8 = %1000_1000
  
                           '!C2 !C1
  RO_NFC_TRAILER_BYTE_6 = %0000_0111 
                           ' C1 !C3
  RO_NFC_TRAILER_BYTE_7 = %1000_1111 
                           ' C3  C2
  RO_NFC_TRAILER_BYTE_8 = %0000_1111 

  NUMBER_OF_KEYS = 16
  
DAT
'' Add serial numbers from previously read cards.

serialNumber            long $739020BC      ' index #0
                        long $138DE88B      ' index #1
                        long END_OF_SERIAL_NUMBER_LIST
                        'long [MAX_POSSIBLE_CARDS - 2]

keyA                    byte $FF[BYTES_PER_KEY * NUMBER_OF_KEYS]
keyB                    byte $FF[BYTES_PER_KEY * NUMBER_OF_KEYS]
madPublicKeyA           byte $A0, $A1, $A2, $A3, $A4, $A5
nfcPublicKeyA           byte $D3, $F7, $D3, $F7, $D3, $F7
newKey                  byte $FF, $FF, $FF, $FF, $FF, $FF, $FF, $07, $80, $69, $FF, $FF, $FF, $FF, $FF, $FF

'read with key A or key B and written with key B
'accessBits = READ_A_B_WRITE_B = %100
'key A is read with key B and never read,
'Access bits is read with key A or key B and written with key B,
'Key B is written with key B and never read.

readWriteAccessBits     byte RW_SECTOR_TRAILER_BYTE_6, RW_SECTOR_TRAILER_BYTE_7, RW_SECTOR_TRAILER_BYTE_8
readOnlyAccessBits      byte RO_SECTOR_TRAILER_BYTE_6, RO_SECTOR_TRAILER_BYTE_7, RO_SECTOR_TRAILER_BYTE_8

readWriteNfcBits     byte RW_NFC_TRAILER_BYTE_6, RW_NFC_TRAILER_BYTE_7, RW_NFC_TRAILER_BYTE_8
readOnlyNfcBits      byte RO_NFC_TRAILER_BYTE_6, RO_NFC_TRAILER_BYTE_7, RO_NFC_TRAILER_BYTE_8

' Access bits setting for MAD sector with read-only access granted.
' 



'' The strings assigned under "correspondingText" need to be in the same order as
'' the order of the serial numbers listed above.
'' Make sure to terminate each string with a zero.
'' Use the method "FindCardIndexNumber" to find the index number of
'' a serial number read from a card. Use this index number with
'' the method "FindStringOrderPtr" to access the strings below.
'' See example in the method "MainLoop".
                             '1234567890123456
correspondingText       byte "Pod Bay Access", 0
                        byte "Password 001", 0
                        'byte 0[(MAX_POSSIBLE_CARDS - 2)]
                       
notFoundText            byte "Unknown Card", 0

DAT

sectorKeyA              byte $FF[18]
sectorNewKeyA           byte $FF[13], $07, $80, $69, $FF[13], $07, $80, $69, $FF[6]

sectorKey               byte $AA[6]
                        byte $BB[6]
                        byte $CC[6]
                        byte $DD[6]
                        byte $EE[6]
                        byte $FF[6]
                        byte $00[6]
                        byte $11[6]
                        byte $22[6]
                        byte $33[6]
                        byte $44[6]
                        byte $55[6]
                        byte $66[6]
                        byte $77[6]
                        byte $88[6]
                        byte $99[6]
                             '1234567890123456
testData                byte "Hello World!", 0[4]
                        byte "Duane Degn", 0 [6]
                        byte "Would you like", 0[2]
                        byte "123456789012345", 0
                             '1234567890123456
numbers                 byte "s:zero b:zero", 0[BLOCK_SIZE - 13]
                        byte "s:zero b:one", 0[BLOCK_SIZE - 12]
                        byte "s:zero b:two", 0[BLOCK_SIZE - 12]
                        byte "s:one b:zero", 0[BLOCK_SIZE - 12]
                        byte "s:one b:one", 0[BLOCK_SIZE - 11]
                        byte "s:one b:two", 0[BLOCK_SIZE - 11]
                        byte "s:two b:zero", 0[BLOCK_SIZE - 13]
                        byte "s:two b:one", 0[BLOCK_SIZE - 13]
                        byte "s:two b:two", 0[BLOCK_SIZE - 13]
                        byte "s:three b:zero", 0[BLOCK_SIZE - 13]
                        byte "s:three b:one", 0[BLOCK_SIZE - 13]
                        byte "s:three b:two", 0[BLOCK_SIZE - 13]
                        byte "s:four b:zero", 0[BLOCK_SIZE - 13]
                        byte "s:four b:one", 0[BLOCK_SIZE - 13]
lastNumberText          byte "s:four b:two", 0[BLOCK_SIZE - 13]
                             '1234567890123456
otherData               byte "zero  Propeller", 0
                        byte "one   Propeller", 0    
                        byte "two   Propeller", 0    
                        byte "three Propeller", 0    
                        byte "four  Propeller", 0    
                        byte "five  Propeller", 0    
                        byte "six   Propeller", 0    
                        byte "seven Propeller", 0    
                        byte "eight Propeller", 0    
                        byte "nine  Propeller", 0
                        byte "zero1 Propeller", 0
                        byte "one1  Propeller", 0    
                        byte "two1  Propeller", 0    
                        byte "three1Propeller", 0    
                        byte "four1 Propeller", 0    
                        byte "five1 Propeller", 0    
                        byte "six1  Propeller", 0    
                        byte "seven1Propeller", 0    
                        byte "eight1Propeller", 0    
                        byte "nine1 Propeller", 0    
                        

'' Each sector has 4 blocks for 16 bytes?                        
                                          