DAT ObjectName         byte "Mfrc522_140623a", 0
CON
{{
  ****** Public Notes ******

  Code to read from MFRC522 RFID sensors.
  
  By: Duane Degn
  For: Dave Dorhout
  
  First attempt uploaded March 20, 2014.
  First working version uploaded March 21, 2014

  Version 140321e appears to read cards and tags
  correctly.

  For updated version of this code see the
  following thread in the Parallax Propeller
  Forum.

  http://forums.parallax.com/showthread.php/154851 
  
}}
{
  ****** Private Notes ******

  21e Change name to "Mfrc522_140321e" in preparation
  to use as stand alone object.
  
}

{{
 Modified from Arduino code. Original code credits below.
 * File name:RFID.pde
 * Creator:Dr.Leong   ( WWW.B2CQSHOP.COM )
 * Creation date:2011.09.19
 * Modified by: Eng. Robson (robson.eletronico@gmail.com)
 * Modified date: 2013.09.10
 * Modified: Translation from Chinese to English (by google)
 * Functional Description:Mifare1 Anti-collision find cards >>> Select card reader interface
 
}}
CON                                                                                                                     

  PCD_IDLE = $00                ''NO action; Cancel the current command
  PCD_AUTHENT = $0E             ''Authentication Key
  PCD_RECEIVE = $08             ''Receive Data
  PCD_TRANSMIT = $04            ''Transmit data
  PCD_TRANSCEIVE = $0C          ''Transmit and receive data,
  PCD_RESETPHASE = $0F          ''Reset
  PCD_CALCCRC = $03             ''CRC Calculate

  PICC_REQIDL = $26             '' find the antenna area does not enter hibernation
  PICC_REQALL = $52             '' find all the cards antenna area
  PICC_ANTICOLL = $93           '' anti-collision
  PICC_SElECTTAG = $93          '' election card
  PICC_AUTHENT1A = $60          '' authentication key A
  PICC_AUTHENT1B = $61          '' authentication key B
  PICC_READ = $30               '' Read Block
  PICC_WRITE = $A0              '' write block
  PICC_DECREMENT = $C0          '' debit
  PICC_INCREMENT = $C1          '' recharge
  PICC_RESTORE = $C2            '' transfer block data to the buffer
  PICC_TRANSFER = $B0           '' save the data in the buffer
  PICC_HALT = $50               '' Sleep

  MAX_LEN = 16                  '' Maximum length of the array
  TEMP_BUFFER_SIZE = MAX_LEN + 2
  
CON

  '' And MF522 The error code is returned when communication
  MI_OK = 0
  MI_NOTAGERR = 1
  MI_ERR = 2


'------------------MFRC522 Register---------------
'Page 0:Command and Status
      Reserved00            = $00    
      CommandReg            = $01    
      CommIEnReg            = $02    
      DivlEnReg             = $03    
      CommIrqReg            = $04    
      DivIrqReg             = $05
      ErrorReg              = $06    
      Status1Reg            = $07
STATUS_1_REG = $07         
      Status2Reg            = $08    
      FIFODataReg           = $09
      FIFOLevelReg          = $0A
      WaterLevelReg         = $0B
      ControlReg            = $0C
      BitFramingReg         = $0D
  BIT_FRAMING_REG = $0D           
      CollReg               = $0E
      Reserved01            = $0F
'Page 1:Command     
      Reserved10            = $10
      ModeReg               = $11
      TxModeReg             = $12
      RxModeReg             = $13
      TxControlReg          = $14
      TxAutoReg             = $15
      TxSelReg              = $16
      RxSelReg              = $17
      RxThresholdReg        = $18
      DemodReg              = $19
      Reserved11            = $1A
      Reserved12            = $1B
      MifareReg             = $1C
      Reserved13            = $1D
      Reserved14            = $1E
      SerialSpeedReg        = $1F
'Page 2:CFG    
      Reserved20            = $20  
      CRCResultRegM         = $21
      CRCResultRegL         = $22
      Reserved21            = $23
      ModWidthReg           = $24
      Reserved22            = $25
      RFCfgReg              = $26
      GsNReg                = $27
      CWGsPReg              = $28
      ModGsPReg             = $29
      TModeReg              = $2A
  T_MODE_REG = $2A      
      TPrescalerReg         = $2B
      TReloadRegH           = $2C
      TReloadRegL           = $2D
      TCounterValueRegH     = $2E
      TCounterValueRegL     = $2F
'Page 3:TestRegister     
      Reserved30            = $30
      TestSel1Reg           = $31
      TestSel2Reg           = $32
      TestPinEnReg          = $33
      TestPinValueReg       = $34
      TestBusReg            = $35
      AutoTestReg           = $36
      VersionReg            = $37
      AnalogTestReg         = $38
      TestDAC1Reg           = $39  
      TestDAC2Reg           = $3A   
      TestADCReg            = $3B   
      Reserved31            = $3C   
      Reserved32            = $3D   
      Reserved33            = $3E   
      Reserved34            = $3F
'-----------------------------------------------
CON

  CHILD_LED = 18
    
VAR

  byte serNum[MAX_LEN]
  byte tempBuffer[TEMP_BUFFER_SIZE]   

  byte spiMosiPin, spiMisoPin, spiClkPin, spiCsPin, resetPin
  
OBJ

  Pst : "Parallax Serial TerminalDat"
  
PUB Init(spiMosiPin_, spiMisoPin_, spiClkPin_, spiCsPin_, resetPin_)

  spiMosiPin := spiMosiPin_
  spiMisoPin := spiMisoPin_
  spiClkPin := spiClkPin_
  spiCsPin := spiCsPin_
  resetPin := resetPin_

  dira[spiClkPin] := 1
  dira[spiMosiPin] := 1
  dira[resetPin] := 1
  outa[resetPin] := 0
  dira[spiCsPin] := 1
  outa[spiCsPin] := 1
  
  ''spiMisoPin remains an input
  
  MfrcInit

PUB GetObjectName

  result := @ObjectName

PUB GetChecksum(localPtr) 

  repeat 4
    result ^= byte[localPtr++]

PUB SwapEndians(distinationPtr, sourcePtr)

  repeat result from 3 to 0
    byte[distinationPtr++] := byte[sourcePtr + result]

PUB WriteMfrc522(registerAddress, value)

  outa[spiCsPin] := 0

  '//Address Format:0XXXXXX0
  SpiTransfer(registerAddress << 1)   
  SpiTransfer(value)
        
  outa[spiCsPin] := 1

PRI SpiTransfer(value)
'' Chip select pin should be control by the method calling this method.

  value ->= 8 ' prepare msb to be rotated into place
  repeat 8
    value <-= 1
    outa[spiMosiPin] := value   ' write to device
    result <<= 1
    outa[spiClkPin] := 1
    result += ina[spiMisoPin]   ' read from device
    outa[spiClkPin] := 0    
   
PUB Mfrc522Write(blockAddress, dataPtr) | receivedBits
'' Write block data
'' Input parameters: blockAddr - block address; writeData - to 16-byte data block write
  
    
  tempBuffer[0] := PICC_WRITE
  tempBuffer[1] := blockAddress
  receivedBits := 0
  CalculateCrc(@tempBuffer, 2, @tempBuffer + 2)
  outa[CHILD_LED] := 1
  outa[CHILD_LED] := 0
  result := Mfrc522ToCard(PCD_TRANSCEIVE, @tempBuffer, 4, @tempBuffer, @receivedBits)

  if result <> MI_OK or receivedBits <> 4 or tempBuffer[0] & $0F <> $0A
    result := MI_ERR   
    outa[CHILD_LED] := 1
    outa[CHILD_LED] := 0
    outa[CHILD_LED] := 1
    outa[CHILD_LED] := 0
    outa[CHILD_LED] := 1
    outa[CHILD_LED] := 0

        
  if result == MI_OK
    bytemove(@tempBuffer, dataPtr, MAX_LEN)
    CalculateCrc(@tempBuffer, MAX_LEN, @tempBuffer + MAX_LEN)
    outa[CHILD_LED] := 1
    outa[CHILD_LED] := 0
    outa[CHILD_LED] := 1
    outa[CHILD_LED] := 0
    result := Mfrc522ToCard(PCD_TRANSCEIVE, @tempBuffer, TEMP_BUFFER_SIZE, @tempBuffer, @receivedBits)
        
    if result <> MI_OK or receivedBits <> 4 or tempBuffer[0] & $0F <> $0A  ' *** ?
      result := MI_ERR   
      outa[CHILD_LED] := 1
      outa[CHILD_LED] := 0
      outa[CHILD_LED] := 1
      outa[CHILD_LED] := 0
      outa[CHILD_LED] := 1
      outa[CHILD_LED] := 0
      outa[CHILD_LED] := 1
      outa[CHILD_LED] := 0
    
  
PUB ReadMfrc522(registerAddress)

  outa[spiCsPin] := 0

  '//Address Format:1XXXXXX0 
  SpiTransfer((registerAddress << 1) | $80)  
  result := SpiTransfer($00) '' single byte result
        
  outa[spiCsPin] := 1
        
PRI SetBitMask(registerAddress, mask)  

  result := ReadMfrc522(registerAddress)
  WriteMfrc522(registerAddress, result | mask)  '// set bit mask

PRI ClearBitMask(registerAddress, mask)   

  result := ReadMfrc522(registerAddress)
  WriteMfrc522(registerAddress, result & !mask) 'clear bit mask
  
PUB AntennaOn

  result := ReadMfrc522(TxControlReg)
  if (result & $03) <> $03 '(!(temp & $03))
    SetBitMask(TxControlReg, $03)

PUB AntennaOff

  ClearBitMask(TxControlReg, $03)

PUB Mfrc522Reset

  WriteMfrc522(CommandReg, PCD_RESETPHASE)

PUB MfrcInit

  outa[resetPin] := 1

  Mfrc522Reset
  
  WriteMfrc522(TModeReg, $8D) ' Timer register
  WriteMfrc522(TPrescalerReg, $3E)
  WriteMfrc522(TReloadRegL, 30)
  WriteMfrc522(TReloadRegH, 0)
  WriteMfrc522(TxAutoReg, $40)
  WriteMfrc522(ModeReg, $3D)
  'ClearBitMask(TPrescalerReg, $3E) ' commented out in Arduino code  
  'WriteMfrc522(RxSelReg, $86)  ' commented out in Arduino code
  'WriteMfrc522(RFCfgReg, $7F)  ' commented out in Arduino code  
  AntennaOn

PUB RequestFromMfrc522(requestMode, tagTypePtr) | backBits
'' Data is sent on rising clock.
'' Data is read 
'' Description: Find cards, read the card type number
'' Input parameters: reqMode - find cards way
''                       TagType - Return Card Type
''                              $4400 = Mifare_UltraLight
''                              $0400 = Mifare_One(S50)
''                              $0200 = Mifare_One(S70)
''                              $0800 = Mifare_Pro(X)
''                              $4403 = Mifare_DESFire
'' Return value: the successful return MI_OK  
 
  backBits := 0
        
  WriteMfrc522(BitFramingReg, $07)             '//TxLastBists = BitFramingReg[2..0]     ???
   
  byte[tagTypePtr][0] := requestMode
  
  result := Mfrc522ToCard(PCD_TRANSCEIVE, tagTypePtr, 1, tagTypePtr, @backBits)
       
  if result <> MI_OK or backBits <> $10
    if backBits <> $10
    result := MI_ERR
                                   
PRI Mfrc522ToCard(command, sendDataPtr, lengthOfDataToSend, backDataPtr, dataBitLenthPtr) | {
} interruptEnable, waitInterrupt, lastBits, quitFlag, tempValue, localIndex, readCounter

  result := MI_ERR
  quitFlag := 0
  interruptEnable := $00
  waitInterrupt := $00
  case command
    PCD_AUTHENT:               '//Certification cards close
      interruptEnable := $12
      waitInterrupt := $10
   
    PCD_TRANSCEIVE:    '//Transmit FIFO data
      interruptEnable := $77
      waitInterrupt := $30
   
  WriteMfrc522(CommIEnReg, interruptEnable | $80)      '//Interrupt request
  ClearBitMask(CommIrqReg, $80)                     '//Clear all interrupt request bit
  SetBitMask(FIFOLevelReg, $80)                     '//FlushBuffer=1, FIFO Initialization
   
  WriteMfrc522(CommandReg, PCD_IDLE)    '//NO action; Cancel the current command???
   
  '//Writing data to the FIFO
  repeat result from 0 to lengthOfDataToSend - 1
    WriteMfrc522(FIFODataReg, byte[sendDataPtr][result])

    '//Execute the command
  WriteMfrc522(CommandReg, command)
  if command == PCD_TRANSCEIVE
    SetBitMask(BitFramingReg, $80)                '//StartSend=1,transmission of data starts  
    '//Waiting to receive data to complete
   '   i = 2000;       //i according to the clock frequency adjustment, the operator M1 card maximum waiting time 25ms???
  repeat readCounter from 0 to 1999 '2000
    '//CommIrqReg[7..0]
    '//Set1 TxIRq RxIRq IdleIRq HiAlerIRq LoAlertIRq ErrIRq TimerIRq
    tempValue := ReadMfrc522(CommIrqReg)
 
    if tempValue & 1 <> 0 or tempValue & waitInterrupt <> 0
      quitFlag := 1
      quit
      
  'while ((i!=0) && !(n&$01) && !(n&waitInterrupt));
   
  ClearBitMask(BitFramingReg, $80)                  '//StartSend=0
      
  if quitFlag
     
    ifnot ReadMfrc522(ErrorReg) & $1B   '//BufferOvfl Collerr CRCErr ProtecolErr      
      result := MI_OK
      
      if tempValue & interruptEnable & $01 ' *** Think about this  
        result := MI_NOTAGERR                  ' //??   
                  
      if command == PCD_TRANSCEIVE
    
        tempValue := ReadMfrc522(FIFOLevelReg)
      
        lastBits := ReadMfrc522(ControlReg) & $07
    
        if lastBits
          byte[dataBitLenthPtr] := ((tempValue - 1) * 8) + lastBits                        
        else          
          byte[dataBitLenthPtr] := tempValue * 8                        
          
        if tempValue == 0   
          tempValue := 1    
           
        if tempValue > MAX_LEN
          tempValue := MAX_LEN                     
                        
        '//Reading the received data in FIFO
        repeat localIndex from 0 to tempValue - 1
          byte[backDataPtr][localIndex] := ReadMfrc522(FIFODataReg)
    
    else    
      result := MI_ERR
  
  '//SetBitMask(ControlReg, $80)           '//timer stops
  '//WriteMfrc522(CommandReg, PCD_IDLE) 

PUB Mfrc522Anticollision(serialNumberPtr) | localIndex, serialNumberCheck, unLen
'' 4 bytes card serial number, the first 5 bytes for the checksum byte
'' successful return MI_OK

  serialNumberCheck := 0    

  '//ClearBitMask(Status2Reg, $08)           '//TempSensclear
  '//ClearBitMask(CollReg, $80)                       '//ValuesAfterColl
  WriteMfrc522(BitFramingReg, $00)             '//TxLastBists = BitFramingReg[2..0]
 
  byte[serialNumberPtr][0] := PICC_ANTICOLL
  byte[serialNumberPtr][1] := $20
  
  result := Mfrc522ToCard(PCD_TRANSCEIVE, serialNumberPtr, 2, serialNumberPtr, @unLen)
 
  if result == MI_OK
    '//Check card serial number
    repeat localIndex from 0 to 3
      serialNumberCheck ^= byte[serialNumberPtr][localIndex]
                
    if serialNumberCheck & $FF <> byte[serialNumberPtr][4]
      result := MI_ERR
            
  '//SetBitMask(CollReg, $80);                //ValuesAfterColl=1

PUB CalculateCrc(dataPtr, length, resultsPtr) | temp

  ClearBitMask(DivIrqReg, $04)                      '//CRCIrq = 0
  SetBitMask(FIFOLevelReg, $80)                     '//Clear the FIFO pointer
  '//WriteMfrc522(CommandReg, PCD_IDLE)
   
  '//Writing data to the FIFO      

  repeat result from 0 to length - 1   
    WriteMfrc522(FIFODataReg, byte[dataPtr][result])   
      
  WriteMfrc522(CommandReg, PCD_CALCCRC)
   
  '//Wait CRC calculation is complete

  repeat result from $FF to 1
    temp := ReadMfrc522(DivIrqReg)  ' ** Can the IRQ pin be monitored.
    if temp == $04
      quit 
  '//CRCIrq = 1
   
  '//Read CRC calculation result
  byte[resultsPtr][0] := ReadMfrc522(CRCResultRegL)
  byte[resultsPtr][1] := ReadMfrc522(CRCResultRegM)

  result := (byte[resultsPtr][1] << 8) + byte[resultsPtr][0]

PUB Mfrc522SelectTag(serialNumberPtr) : size | status, receivedBits
'' election card, read the card memory capacity
'' Presently not used.

  '//ClearBitMask(Status2Reg, $08)                       //MFCrypto1On=0

  tempBuffer[0] := PICC_SElECTTAG
  tempBuffer[1] := $70
  receivedBits := 0
  bytemove(@tempBuffer + 2, serialNumberPtr, 5)
    
  CalculateCrc(@tempBuffer, 7, @tempBuffer + 7)        '     //??
  status := Mfrc522ToCard(PCD_TRANSCEIVE, @tempBuffer, 9, @tempBuffer, @receivedBits)
    
  if status == MI_OK and receivedBits == $18
    size := tempBuffer[0] 
  else
    size := 0    
  
PUB MfrcAuthentication(authenticationMode, blockAddress, sectorKeyPtr, serialNumberPtr) | {
} receivedBits, readResult
'' Verify the command block address + sector + password + card serial number
'' Presently not used.

  tempBuffer[0] := authenticationMode
  tempBuffer[1] := blockAddress
  receivedBits := 0
  bytemove(@tempBuffer + 2, sectorKeyPtr, 6)      
  bytemove(@tempBuffer + 8, serialNumberPtr, 4)      

  result := Mfrc522ToCard(PCD_AUTHENT, @tempBuffer, 12, @tempBuffer, @receivedBits)

  'if result <> MI_OK or ReadMfrc522(Status2Reg) & $08 == 0
  readResult := ReadMfrc522(Status2Reg)
  
  if result <> MI_OK or readResult & $08 == 0
    Pst.Str(string(13, "MfrcAuthentication Error"))
    Pst.Str(string(13, "Mfrc522ToCard = "))
    Pst.Dec(result)
    Pst.Str(string(13, "ReadMfrc522 = $"))
    Pst.Hex(readResult, 2)
    
    result := MI_ERR

''uchar MFRC522_Auth(uchar authMode, uchar BlockAddr, uchar *Sectorkey, uchar *serNum)
{
    uchar status;
    uint recvBits;
    uchar i;
    uchar buff[12]; 

        //Verify the command block address + sector + password + card serial number
    buff[0] = authMode;
    buff[1] = BlockAddr;
    for (i=0; i<6; i++)
    {    
                buff[i+2] = *(Sectorkey+i);   
        }
    for (i=0; i<4; i++)
    {    
                buff[i+8] = *(serNum+i);   
        }
    status = MFRC522_ToCard(PCD_AUTHENT, buff, 12, buff, &recvBits);

    if ((status != MI_OK) || (!(Read_MFRC522(Status2Reg) & 0x08)))
    {   
                status = MI_ERR;   
        }
    
    return status;
}
PUB Mfrc522Read(blockAddress, dataPtr) | unLen
'' Presently not used.

  byte[dataPtr][0] := PICC_READ
  byte[dataPtr][1] := blockAddress
  CalculateCrc(dataPtr, 2, dataPtr + 2)
  result := Mfrc522ToCard(PCD_TRANSCEIVE, dataPtr, 4, dataPtr, @unLen)
   
  if ((result <> MI_OK) or (unLen <> $90))
    result := MI_ERR
   

PUB Mrfc522Hibernate | unLen
'' Command card into hibernation
 
    tempBuffer[0] := PICC_HALT
    tempBuffer[1] := 0
    CalculateCrc(@tempBuffer, 2, @tempBuffer + 2)
 
    result := Mfrc522ToCard(PCD_TRANSCEIVE, @tempBuffer, 4, @tempBuffer, @unLen)
                                    